@@ -1,4 +1,7 @@
 /*
+ * Like what you see? Join us!
+ * https://www.univention.com/about-us/careers/vacancies/
+ *
  * Copyright 2011-2022 Univention GmbH
  *
  * https://www.univention.de/
@@ -248,7 +251,7 @@ define([
 							return;
 						}
 
-						if (limit === this._grid.getAllItems().length) {
+						if (limit <= this._grid.getAllItems().length) {
 							var content = _('The search query has reached the limit of %s entries. The limit can be adjusted with the UCR variable <code>admin/diary/query/limit</code>.', limit);
 							this._queryLimitWarning.set('content', content);
 							this._queryLimitWarning.set('visible', true);
