@@ -4,6 +4,9 @@
 # Univention Management Console
 #  module: manages system services
 #
+# Like what you see? Join us!
+# https://www.univention.com/about-us/careers/vacancies/
+#
 # Copyright 2011-2022 Univention GmbH
 #
 # https://www.univention.de/
@@ -45,7 +48,7 @@ import univention.management.console as umc
 from univention.management.console.log import MODULE
 
 from univention.admindiary.client import add_comment
-from univention.admindiary.backend import get_client, get_engine, get_query_limit
+from univention.admindiary.backend import get_client, get_engine, get_query_limit, NoDBConnection
 from univention.admindiary.events import DiaryEvent
 
 
@@ -107,8 +110,11 @@ class Instance(Base):
 
 	@simple_response
 	def options(self):
-		with get_client(version=1) as client:
-			return client.options()
+		try:
+			with get_client(version=1) as client:
+				return client.options()
+		except NoDBConnection as exc:
+			raise umcm.UMC_Error(str(exc), status=500)
 
 	@simple_response
 	def get(self, context_id):
